#!/usr/bin/python3
# coding: utf-8

"""
Plot graphs with data given by operations

This script receives a file path as an argument and outputs, in the same folder
as where the input is located, a bunch of plots

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2022
"""

import pandas as pd
import datetime
import matplotlib.pyplot as plt
import sys

def number_of_nodes(trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Histogram of the number of nodes asked for each job
    :param trace: Input pandas Data Frame with data from REAL slurm data
    :param file_name: Original data file name
    :param dir_path: Directory where the plot will be saved
    :return: None
    """
    print(f"Plot number of requested nodes histogram...")
    plt.plot(trace["NNodes"], ".")
    plt.title(f"Distribution of Nodes requested\n obs_trace from {file_name}")
    plt.ylabel("Number of Nodes")
    plt.xlabel("Number of Jobs")
    plt.savefig(f"{dir_path}/{file_name}_nodes.png")
    plt.clf()

def processors(trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Histogram of the number of processors asked for each job
    :param trace: Input pandas Data Frame with data from REAL slurm data
    :param file_name: Original data file name
    :param dir_path: Directory where the plot will be saved
    :return: None
    """
    print(f"Plot number of requested cpus histogram...")
    plt.hist(trace["NCPUS"])
    plt.title(f"Distribution of CPUs requested\n Data from {file_name}")
    plt.ylabel("Number of CPUs")
    plt.xlabel("Number of Jobs")
    plt.savefig(f"{dir_path}/{file_name}_cpus.png")
    plt.clf()

def processors_box_plot(trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    print(f"Plot box plot of requested cpus...")
    # without outliers
    plt.boxplot(trace["NCPUS"], labels=[file_name], showfliers=False)
    plt.title(f"Box plot of requested CPUS")
    plt.ylabel("Number of CPUs")
    plt.savefig(f"{dir_path}/{file_name}_box_plot_cpus.png")
    plt.clf()

def nodes_box_plot(trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    print(f"Plot box plot of requested nodes...")
    # without outliers
    plt.boxplot(trace["NNodes"], labels=[file_name])
    plt.title(f"Box plot of requested nodes")
    plt.ylabel("Number of Nodes")
    plt.savefig(f"{dir_path}/{file_name}_box_plot_nodes.png")
    plt.clf()

def cpus_times_elapsed(trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    print(f"Requested CPUs by elapsed time in seconds...")
    trace_without_pending = trace[trace["Elapsed in seconds"] != 0]
    plt.plot(trace_without_pending["Elapsed in seconds"], trace_without_pending["NNodes"], '.')
    plt.title(f"Requested CPUs per elapsed time\nOnly considering jobs that started {file_name}")
    plt.ylabel("Number of CPUs")
    plt.xlabel("Elapsed time in seconds")
    plt.savefig(f"{dir_path}/{file_name}_cpu_x_elapsed.png")
    plt.clf()

def submit_time(trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Histogram of the start time with respect to the first started job
    :param trace: Input pandas Data Frame with data from REAL slurm data
    :param file_name: Original data file name
    :param dir_path: Directory where the plot will be saved
    :return: None
    """
    print(f"Plot submit time histogram...")
    plt.hist(trace["Zeroed submit"])
    plt.title(f"Distribution of submission time requested\n wrt to the first submission. Data from {file_name}")
    plt.ylabel("Number of Jobs")
    plt.xlabel("Submit time in seconds wrt to the first submission")
    plt.savefig(f"{dir_path}/{file_name}_submit_time.png")
    plt.clf()


def state_jobs(trace: pd.DataFrame, file_name: str, dir_path: str) -> None:
    """
    Pie chart of the states of the jobs
    :param trace: Input pandas Data Frame with data from REAL slurm data
    :param file_name: Original data file name
    :param dir_path: Directory where the plot will be saved
    :return: None
    """
    print(f"Plot pie-chart with job status...")
    states = trace["State"].unique()
    num_per_state = []
    for state in states:
        num_per_state.append(len(trace[trace["State"] == state]))
    plt.pie(num_per_state, labels=states)
    plt.title(f"Number of jobs per state\nData from {file_name}")
    plt.savefig(f"{dir_path}/{file_name}_states.png")
    plt.clf()


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("You have to provide the path to the CSV with COMMA AS DELIMITER")
        sys.exit(1)

    file_path = sys.argv[1]
    data_name = file_path.split("/")[-1].split(".")[0]  # picks the file name without path and file extension
    data_path = file_path.rsplit("/", 1)[0]  # remove the file name from directory path
    obs_trace = pd.read_fwf(file_path, infer_nrows=2)

    print(f"Preprocessing {data_name}")

    if "-" in obs_trace.iloc[0]["JobID"]:
        # remove the line with the width information
        obs_trace.drop(0, inplace=True)

    # here we will fill the subjob's parameters with the parent job
    last_job = None # we keep track of the last job
    for i, row in obs_trace.iterrows():
        if row.isnull().any():
            # then this is a subjob, then we complete with the previous job
            for col_index, is_not_filled in enumerate(row.isnull()):
                if is_not_filled:
                    obs_trace.at[i-1, obs_trace.columns[col_index]] = last_job[obs_trace.columns[col_index]]
        else:
            # this is a job
            last_job = row

    # transforming to integer values
    obs_trace[["NNodes", "NCPUS", "Priority"]] = obs_trace[["NNodes", "NCPUS", "Priority"]].apply(pd.to_numeric)

    # Transform submit data to datetime to make it easier to compute date differences
    formatted_date = []
    for i in range(len(obs_trace["Submit"])):
        formatted_date.append(datetime.datetime.strptime(obs_trace.iloc[i]["Submit"], "%Y-%m-%dT%H:%M:%S"))
    # 8 is the position of the column, just for aesthetics 
    obs_trace.insert(8, "Formatted submit", formatted_date, allow_duplicates=True)

    print(f"Add formatted submit column")

    # this array will be the submitted time wrt reference in the first submitted job
    reference = min(formatted_date)  # pick the first job as the reference
    zeroed_date = []
    # We take as the reference the submission of the first job
    for i in range(len(obs_trace["Formatted submit"])):
        # And take the time in seconds wrt to the reference
        zeroed_date.append((obs_trace.iloc[i]["Formatted submit"]-reference).seconds)
    obs_trace.insert(9, "Zeroed submit", zeroed_date, allow_duplicates=True)

    print(f"Add zeroed submit column")

    # Transform elapsed format to seconds
    elapsed_in_seconds = []
    for i in range(len(obs_trace["Elapsed"])):
        current_data = obs_trace.iloc[i]["Elapsed"].split(":")
        elapsed_in_seconds.append((int(current_data[0])*60 + int(current_data[1]))*60 + int(current_data[2]))
    obs_trace.insert(7, "Elapsed in seconds", elapsed_in_seconds, allow_duplicates=True)

    number_of_nodes(obs_trace, data_name, data_path)
    nodes_box_plot(obs_trace, data_name, data_path)
    submit_time(obs_trace, data_name, data_path)
    processors(obs_trace, data_name, data_path)
    processors_box_plot(obs_trace, data_name, data_path)
    state_jobs(obs_trace, data_name, data_path)
    cpus_times_elapsed(obs_trace, data_name, data_path)
