#! /usr/bin/python3

"""
Converts a sacct output to Standard Workflow Format[1].

This script will anonymize uid and gid.

[1] Chapin, S.J. et al. (1999). Benchmarks and Standards for the Evaluation of Parallel Job Schedulers.
In: Feitelson, D.G., Rudolph, L. (eds) Job Scheduling Strategies for Parallel Processing. JSSPP 1999.
Lecture Notes in Computer Science, vol 1659. Springer, Berlin, Heidelberg.

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2022
"""

import sys
import os.path
import pandas as pd
from datetime import timedelta

def strfseconds(input:str):
    """
    Reads a string formatted as [D-]HH:MM:SS and returns the time in seconds, if possible. Otherwise -1.
    :param input: formatted string
    :return: either the total in seconds or -1
    """
    days = 0
    try:
        split_hyphen = input.split("-")
    except AttributeError:
        # this is not a string, we return
        return -1
    if len(split_hyphen) == 2:
        try:
            days = int(split_hyphen[0])
        except ValueError:
            # the format is not the expected one if we cannot cast the left part of the hyphen to an integer
            return -1
        time = split_hyphen[1]
    else:
        time = split_hyphen[0]

    split_dots = time.split(":")

    if len(split_dots) != 3:
        # if there are not 3 values if we split at :, then we are not in the expected format
        return -1

    try:
        hours = int(split_dots[0])
        minutes = int(split_dots[1])
        seconds = int(split_dots[2])
    except ValueError:
        # if we fail to cast to integer any of the split values, then we are out of the expected format
        return -1

    return days*3600*24 + hours * 3600 + minutes * 60 + seconds

def strfkbs(input:str):
    """
    reads a string as formatted by acct and returns an integer representing it in kilobytes
    :param input: string as formated by sacct with memory usage
    :return: memory in kilobytes
    """
    if "G" in input:
        try:
            gigabytes = int(input.strip("G"))
        except ValueError:
            # couldn't cast, then there is some gibberish in the field
            return -1
        # 1024*1024*gigabytes
        return 1048576*gigabytes

    if "M" in input:
        try:
            megabytes = int(input.strip("M"))
        except ValueError:
            # couldn't cast, then there is some gibberish in the field
            return -1
        return 1024*megabytes

    if "K" in input:
        try:
            kilobytes = int(input.strip("K"))
        except ValueError:
            # couldn't cast, then there is some gibberish in the field
            return -1
        return kilobytes

    # If it doesn't enter any of the previous cases then it is 0n, which is the null usage recording of memory, or gibberish
    return -1

if __name__ == "__main__":
    # ENVIRONMENT vars
    cur_path = os.path.curdir

    # SCRIPT INPUT
    if len(sys.argv) == 1:
        sys.exit("You must provide a sacct output file to be processed!\nUSAGE: ./slurm2swf.py inputfile output.swf")
    elif len(sys.argv) == 2:
        # input file name
        trace_filename = sys.argv[1]
        # DEFAULT OUTPUTNAME
        swf_filename = "output.swf"
    # output file name
    else:
        trace_filename = sys.argv[1]
        swf_filename = sys.argv[1]

    # FILE INPUT/OUTPUT

    if os.path.isfile(trace_filename):
        # read to pandas using fixed width format. We only require the first 2 rows to infer the length of each of the fields
        sacct_trace = pd.read_fwf(trace_filename, infer_nrows=2)
    else:
        sys.exit(f"sacct output file {trace_filename} does not exist.")

    # FILTERING
    # get rid of the line with hyphens
    # checking if the line is COMPLETE out of hyphens, otherwise do nothing
    appended_values = "".join(sacct_trace.iloc[0])
    if len(appended_values) == appended_values.count("-"):
        sacct_trace.drop(0, inplace=True)

    # get rid of steps. If the jobid has a dot, then we know that it is part of the original's id
    sacct_trace = sacct_trace[~sacct_trace["JobID"].str.contains(".", regex=False)]

    # convert to numeric cpu count fields
    sacct_trace[["AllocCPUS","ReqCPUS"]] = sacct_trace[["AllocCPUS", "ReqCPUS"]].apply(pd.to_numeric, downcast='integer')

    # only real number column. We do cast it only if it is not completely empty
    if not sacct_trace["AveCPU"].isna().all():
        sacct_trace[["AveCPU"]] = sacct_trace[["AveCPU"]].apply(pd.to_numeric)

    # convert to date the appropriate fields: submit, waittime, elapsed, timelimit
    sacct_trace["Submit"] = pd.to_datetime(sacct_trace["Submit"], format="%Y-%m-%dT%H:%M:%S")
    sacct_trace["Start"] = pd.to_datetime(sacct_trace["Start"], format="%Y-%m-%dT%H:%M:%S")

    # create a new column with the wait time
    sacct_trace["Waittime"] = sacct_trace['Start'] - sacct_trace['Submit']

    # convert timedate to time delta
    for i, row in sacct_trace.iterrows():
        elapsed_seconds = strfseconds(row["Elapsed"])
        timelimit_seconds = strfseconds(row["Timelimit"])
        if elapsed_seconds != -1:
            sacct_trace.at[i, "Elapsed"] = timedelta(seconds=elapsed_seconds)
        if timelimit_seconds != -1:
            sacct_trace.at[i, "Timelimit"] = timedelta(seconds=timelimit_seconds)
        else:
            sacct_trace.at[i, "Timelimit"] = -1

    # keep track of the start and finish of the simulation's jobs
    start_time = sacct_trace['Submit'].min()
    end_time = (sacct_trace['Submit'] + sacct_trace['Elapsed']).max()
    for i, row in sacct_trace.iterrows():
        # finally, transform submit to elapsed seconds since the start of the simulation
        sacct_trace.at[i, "Submit"] = (sacct_trace.at[i, "Submit"] - start_time).total_seconds()

    # Memory convert
    for i, row in sacct_trace.iterrows():
        sacct_trace.at[i, "ReqMem"] = strfkbs(row["ReqMem"])

    # SWF status codes
    for i, row in sacct_trace.iterrows():
        if "COMPLETED" in row["State"]:
            sacct_trace.at[i, "State"] = 0
        elif "CANCELLED" in row["State"]:
            sacct_trace.at[i, "State"] = 5
        else:
            sacct_trace.at[i, "State"] = -1

    # we factorize to anonymize data. We get rid of the index table
    sacct_trace["JobID"] = pd.factorize(sacct_trace["JobID"])[0]
    sacct_trace["UID"] = pd.factorize(sacct_trace["UID"])[0]
    sacct_trace["GID"] = pd.factorize(sacct_trace["GID"])[0]

    # queues and partitions we factorize, but keep the dictionary since they carry hard restrictions on the usage
    sacct_trace["QOS"], queues = pd.factorize(sacct_trace["QOS"])
    sacct_trace["Partition"], partitions = pd.factorize(sacct_trace["Partition"])

    # cast, again, dates to integer. Its final form
    for i, row in sacct_trace.iterrows():
        sacct_trace.at[i, "Waittime"] = row["Waittime"].total_seconds()
        sacct_trace.at[i, "Elapsed"] = row["Elapsed"].total_seconds()
        if row["Timelimit"] != -1:
            sacct_trace.at[i, "Timelimit"] = row["Timelimit"].total_seconds()

    sacct_trace[["Waittime", "Elapsed", "Timelimit"]] = sacct_trace[["Waittime", "Elapsed", "Timelimit"]].apply(
        pd.to_numeric, downcast='integer')

    # finally, set to -1 all empty fields
    sacct_trace.fillna(-1, inplace=True)

    # find out the size of each of the fields
    jobid_width = len(str(sacct_trace["JobID"].max()))
    submit_width = len(str(sacct_trace["Submit"].max()))
    # absolute here just because in simulations the difference might be negative :/
    waittime_width = len(str(sacct_trace["Waittime"].max()))
    if len(str(sacct_trace["Waittime"].min())) > waittime_width:
        waittime_width = len(str(sacct_trace["Waittime"].min()))
    elapsed_width = len(str(sacct_trace["Elapsed"].max()))
    alloccpus_width = len(str(sacct_trace["AllocCPUS"].max()))
    avecpu_width = len(str(sacct_trace["AveCPU"].max()))
    reqcpus_width = len(str(sacct_trace["ReqCPUS"].max()))
    timelimit_width = len(str(sacct_trace["Timelimit"].max()))
    reqmem_width = len(str(sacct_trace["ReqMem"].max()))
    # We check if -1 is in the values of state to see the maximum width it can have
    if -1 in sacct_trace["State"].values:
        state_width = 2
    else:
        state_width = 1
    uid_width = len(str(sacct_trace["UID"].max()))
    gid_width = len(str(sacct_trace["GID"].max()))
    qos_width = len(str(sacct_trace["QOS"].max()))
    partition_width = len(str(sacct_trace["Partition"].max()))

    # TODO: OUTPUT TO A FIXED WIDTH FORMAT
    with open(swf_filename, "w") as swf_output:
        # HEADER
        swf_output.write("; Version: 2.2\n")
        swf_output.write("; Computer: \n")
        swf_output.write("; Installation: \n")
        swf_output.write("; Acknowledge: \n")
        swf_output.write(";\n")
        swf_output.write(f"; Conversion: Automatically generated by sacct2swf\n")
        swf_output.write(f"; Conversion: Manuel G. Marciani @ BSC-CNS Earth Sciences \n")
        swf_output.write(";\n")
        swf_output.write("; UnixStartTime: \n")
        swf_output.write("; TimeZoneString: \n")
        swf_output.write(f"; StartTime: {start_time}\n")
        swf_output.write(f"; EndTime: {end_time}\n")
        swf_output.write(";\n")
        swf_output.write(f"; MaxJobs: {len(sacct_trace)}\n")
        swf_output.write(f"; MaxRecords: {len(sacct_trace)}\n")
        # MACHINE PARAMETERS
        swf_output.write("; MaxNodes: \n")
        swf_output.write("; MaxProcs: \n")
        swf_output.write("; Preemption: \n")
        swf_output.write(";\n")
        # QUEUES
        swf_output.write("; Queues: ----------- \n")
        swf_output.write("; Queues:  id   name  \n")
        swf_output.write("; Queues: ----------- \n")
        for queue_id, queue_name in enumerate(queues):
            swf_output.write(f"; Queue: {queue_id}   {queue_name}\n")
        swf_output.write(f"; MaxQueues: {len(queues)}\n")
        swf_output.write(";\n")
        # PARTITIONS
        swf_output.write("; Partitions: ----------- \n")
        swf_output.write("; Partitions:  id   name  \n")
        swf_output.write("; Partitions: ----------- \n")
        for partitions_id, partitions_name in enumerate(partitions):
            swf_output.write(f"; Partition: {partitions_id}   {partitions_name}\n")
        swf_output.write(f"; MaxPartitions: {len(partitions)}\n")
        swf_output.write(";\n")
        for job_id, job in sacct_trace.iterrows():
            jobid_string = "{0:>{width}}".format(job['JobID'], width=jobid_width)
            submit_string = "{0:>{width}}".format(int(job['Submit']), width=submit_width)
            waittime_string = "{0:>{width}}".format(job['Waittime'], width=waittime_width)
            elapsed_string = "{0:>{width}}".format(job['Elapsed'], width=elapsed_width)
            alloccpus_string = "{0:>{width}}".format(job['AllocCPUS'], width=alloccpus_width)
            avecpu_string = "{0:>{width}}".format(job['AveCPU'], width=avecpu_width)
            reqcpus_string = "{0:>{width}}".format(job['ReqCPUS'], width=reqcpus_width)
            timelimit_string = "{0:>{width}}".format(job['Timelimit'], width=timelimit_width)
            reqmem_string = "{0:>{width}}".format(job['ReqMem'], width=reqmem_width)
            state_string = "{0:>{width}}".format(job['State'], width=state_width)
            uid_string = "{0:>{width}}".format(job['UID'], width=uid_width)
            gid_string = "{0:>{width}}".format(job['GID'], width=gid_width)
            qos_string = "{0:>{width}}".format(job['QOS'], width=qos_width)
            partition_string = "{0:>{width}}".format(job['Partition'], width=partition_width)

            swf_output.write(f"{jobid_string} {submit_string} {waittime_string} {elapsed_string} {alloccpus_string} "
                             f"{avecpu_string} -1 {reqcpus_string} {timelimit_string} {reqmem_string} {state_string} "
                             f"{uid_string} {gid_string} -1 {qos_string} {partition_string} -1 -1\n")
