#!/usr/bin/env python3
"""
This is python parser for the SWF format.

Standard Workflow Format specification can be found at https://www.cs.huji.ac.il/labs/parallel/workload/swf.html.

[1] Chapin, S.J. et al. (1999). Benchmarks and Standards for the Evaluation of Parallel Job Schedulers.
In: Feitelson, D.G., Rudolph, L. (eds) Job Scheduling Strategies for Parallel Processing. JSSPP 1999.
Lecture Notes in Computer Science, vol 1659. Springer, Berlin, Heidelberg.

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2022
"""

import datetime # just for computing the difference between the end of the simulation and the start in seconds

def parse_time(value) -> datetime.datetime:
    """
    This function parses a date as given by the output of "date" program in Linux.
    :param value: string with the date
    :return: datetime object with the parsed date
    """
    try:
        return datetime.datetime.strptime(value, "%a %b %d %H:%M:%S %Z %Y")
    except ValueError:
        print(f"[WARNING] {value} failed to create valid datetime object for start time. Ignoring timezone name.")
        split_value = value.split()
        if len(split_value) == 6:
            value = " ".join(split_value[:4] + [split_value[5]])  # remove timezone name
            try:
                return datetime.datetime.strptime(value, "%a %b %d %H:%M:%S %Y")
            except ValueError:
                print(f"[WARNING] {value} is not in the expected date format without timezone (e.g. Mon Nov 17 14:01:47 2008). Ignoring it.")
        else:
            print(f"[WARNING] {value} is not in the expected date format for start time (e.g. Mon Nov 17 14:01:47 CET 2008). Ignoring it.")

class Job:
    """
    Represents a job.
    """

    def __init__(self, job_id, submit_time, wait_time, runtime, alloc_cpus, avg_cpu_used, used_memory, requested_cpus,
                 requested_time, requested_memory, status, uid, gid, app_id, qos, partition, preceding, think_time) -> None:
        """
        Constructor of a single Job from a sequence of parameters
        :param job_id: counter starting from 1.
        :param submit_time: submit time in SECONDS since the start of the simulation.
        :param wait_time: wait time in seconds of the job (this is used for real traces).
        :param runtime: run time in seconds of the job.
        :param alloc_cpus: number of processors that a job uses (useful for real traces).
        :param avg_cpu_used: average use of cpus (real traces).
        :param used_memory: kilobytes of memory used. Average per cpu.
        :param requested_cpus: number of cpus requested.
        :param requested_time: either runtime or average cpu time per processor.
        :param requested_memory: kilobytes of requested memory per processor.
        :param status: status of the job (real traces).
        :param uid: User ID associated with username.
        :param gid: Group ID associated with UID.
        :param app_id: id of the application. Jobs can be bundled under the same application.
        :param qos: queue used.
        :param partition: partition number that it is used.
        :param preceding: id of the job dependency. # manuel: nor sure how this works in slurm simulator
        :param think_time: time elapsed between termination of the preceding job and submittal of the current.
        """

        self._job_id = job_id
        self._submit_time = submit_time
        self._wait_time = wait_time
        self._runtime = runtime
        self._alloc_cpus = alloc_cpus
        self._avg_cpu_used = avg_cpu_used
        self._used_memory = used_memory
        self._requested_cpus = requested_cpus
        self._requested_time = requested_time
        self._requested_memory = requested_memory
        self._status = status
        self._uid = uid
        self._gid = gid
        self._app_id = app_id
        self._qos = qos
        self._partition = partition
        self._preceding = preceding
        self._think_time = think_time

    @property
    def job_id(self):
        """
        :return: Job ID
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        self._job_id = job_id

    @property
    def submit_time(self):
        """
        :return: Submit time in seconds from the start of the simulation.
        """
        return self._submit_time

    @submit_time.setter
    def submit_time(self, submit_time):
        self._submit_time = submit_time

    @property
    def wait_time(self):
        """
        :return: Difference between the time the job was submitted and the tmie it actually began to run.
        """
        return self._wait_time

    @wait_time.setter
    def wait_time(self, wait_time):
        self._wait_time = wait_time

    @property
    def runtime(self):
        """
        :return: end time minus start time of the job.
        """
        return self._runtime

    @runtime.setter
    def runtime(self, runtime):
        self._runtime = runtime

    @property
    def alloc_cpus(self):
        """
        :return: number of allocated processors. In most cases this is also the number of processors the job uses.
        """
        return self._alloc_cpus
    @alloc_cpus.setter
    def alloc_cpus(self, alloc_cpus):
        self._alloc_cpus = alloc_cpus

    @property
    def avg_cpu_used(self):
        """
        :return: average
        """
        return self._avg_cpu_used

    @avg_cpu_used.setter
    def avg_cpu_used(self, avg_cpu_used):
        self._avg_cpu_used = avg_cpu_used

    @property
    def used_memory(self):
        return self._used_memory

    @used_memory.setter
    def used_memory(self, used_memory):
        self._used_memory = used_memory

    @property
    def requested_cpus(self):
        return self._requested_cpus

    @requested_cpus.setter
    def requested_cpus(self, requested_cpus):
        self._requested_cpus = requested_cpus

    @property
    def requested_time(self):
        return self._requested_time

    @requested_time.setter
    def requested_time(self, requested_time):
        self._requested_time = requested_time

    @property
    def requested_memory(self):
        return self._requested_memory

    @requested_memory.setter
    def requested_memory(self, requested_memory):
        self._requested_memory = requested_memory

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, status):
        self._status = status

    @property
    def uid(self):
        return self._uid

    @uid.setter
    def uid(self, uid):
        self._uid = uid

    @property
    def gid(self):
        return self._gid

    @gid.setter
    def gid(self, gid):
        self._gid = gid

    @property
    def app_id(self):
        return self._app_id

    @app_id.setter
    def app_id(self, app_id):
        self._app_id = app_id

    @property
    def qos(self):
        return self._qos

    @qos.setter
    def qos(self, qos):
        self._qos = qos

    @property
    def partition(self):
        return self._partition

    @partition.setter
    def partition(self, partition):
        self._partition = partition

    @property
    def preceding(self):
        return self._preceding

    @preceding.setter
    def preceding(self, preceding):
        self._preceding = preceding

    @property
    def think_time(self):
        return self._think_time

    @think_time.setter
    def think_time(self, think_time):
        self._think_time = think_time

    def to_bsc_swf(self) -> str:
        """
        :return: return the job representation in swf format as in the bsc slurm simulator. They use commas instead of white spaces.
        """
        output = f"{self.job_id},{self.submit_time},{self.wait_time},{self.runtime},{self.alloc_cpus}," \
                 f"{self.avg_cpu_used},{self.used_memory},{self.requested_cpus},{self.requested_time}," \
                 f"{self.requested_memory},{self.status},{self.uid},{self.gid},{self.app_id}," \
                 f"{self.qos},{self.partition},{self.preceding},{self.think_time}\n"
        return output

class SWFtrace:
    """
    This class parses an .swf file and calls swf2trace to generate the binary trace file that sim_mgr runs.
    """

    # generate a swf representation from sequence
    def __init__(self, swf_version, computer_name, location, acknowledge, conversion, max_jobs, max_records, preemption,
                 unix_start_time, time_zone_string, start_time, end_time, max_nodes, max_procs, max_runtime, max_memory, allow_overuse,
                 max_queues, queues, queue, max_partitions, partitions, partition, note, job_list, file_location, avg_cpu, var_cpu,
                 avg_job_rate, unique_uid, unique_gid):
        """
        Represents a SWF file.
        :param swf_version: SWF version of the file
        :param computer_name: name of the represented platform
        :param location: place/institute/center where the platform is housed
        :param acknowledge: name of the person who provided the trace file
        :param conversion: name of the person that converted the trace to SWF format
        :param max_jobs: total number of jobs in the trace
        :param max_records: total number of records in the trace
        :param preemption: Enumerated, with four possible values. 'No' means that jobs run to completion, and are
        represented by a single line in the file. 'Yes' means that the execution of a job may be split into several
        parts, and each is represented by a separate line. 'Double' means that jobs may be split, and their information
        appears twice in the file: once as a one-line summary, and again as a sequence of lines representing the parts,
        as suggested above. 'TS' means time slicing is used, but no details are available
        :param unix_start_time: When the log starts in Unix time
        :param time_zone_string: indicates the time zone which the log was generated
        :param start_time: start time of the log in human-readable form
        :param end_time: end time in human-readable form
        :param max_nodes: number of nodes in the computer
        :param max_procs: number of processors in the computer
        :param max_runtime: maximum runtime that the system allows
        :param max_memory: maximum memory that the system allows
        :param allow_overuse: boolean. "yes" if a job may use more that it requested for any resource. "no", otherwise.
        :param max_queues: number of queues in the system
        :param queues: verbal description of each of the queues (that is their limits)
        :param max_partitions: number of partitions in the system
        :param partitions: verbal description of each of the partitions (that is their limits)
        :param partition: description of a single partition
        :param note: extra info not encapsulated in the fields.
        :param avg_cpu: average of the requested cpus
        :param var_cpu: variance of the requested cpus
        :param avg_job_rate: average number of jobs per seconds
        :param unique_uid: list will all the different uids
        :param unique_gid: list with all the different gids
        """
        # SWF specification
        self._swf_version = swf_version
        self._computer = computer_name
        self._location = location
        self._acknowledge = acknowledge
        self._conversion = conversion
        self._max_jobs = max_jobs
        self._max_records = max_records
        self._preemption = preemption
        self._unix_start_time = unix_start_time
        self._time_zone_string = time_zone_string
        self._start_time = start_time
        self._end_time = end_time
        self._max_nodes = max_nodes
        self._max_procs = max_procs
        self._max_runtime = max_runtime
        self._max_memory = max_memory
        self._allow_overuse = allow_overuse
        self._max_queues = max_queues
        self._queues = queues
        self._queue = queue
        self._max_partitions = max_partitions
        self._partitions = partitions
        self._partition = partition
        self._note = note

        # internal parameters
        self._job_list = job_list
        self._output = file_location

        # statistics
        self._avg_cpu = avg_cpu # average number of allocated cpus
        self._var_cpu = var_cpu # variance of the cpu
        self._avg_job_rate = avg_job_rate
        self._unique_uid = unique_uid
        self._unique_gid = unique_gid

    @classmethod
    def from_file(cls, input_file):
        swf_version = None
        computer_name = None
        location = None
        acknowledge = None
        conversion = None
        max_jobs = None
        max_records = None
        preemption = None
        unix_start_time = None
        time_zone_string = None
        start_time = None
        end_time = None
        max_nodes = None
        max_procs = None
        max_runtime = None
        max_memory = None
        allow_overuse = None
        max_queues = None
        queues = None
        queue = None
        max_partitions = None
        partitions = None
        partition = None
        note = None
        job_list = []
        # statistics
        avg_cpu = 0
        var_cpu = 0
        avg_job_rate = 0
        unique_uid = []
        unique_gid = []

        with open(input_file, "r") as file:
            for line_num, line in enumerate(file):
                if line[0] == ";":
                    # this is a comment line. Then we remove the first character.
                    # TODO there is a more elegant way to do this in python.

                    try:
                        field, value = line[1:].rstrip("\n").split(": ")
                    except ValueError:
                        # this is an empty comment line, so we do nothing
                        continue

                    # remove white spaces
                    field = field.lstrip().rstrip()
                    value = value.rstrip().lstrip()
                    if field == "Version":
                        swf_version = value
                    elif field == "Computer":
                        computer_name = value
                    elif field == "Installation":
                        location = value
                    elif field == "Information":
                        information = value
                    elif field == "Acknowledge":
                        acknowledge = value
                    elif field == "Conversion":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        conversion = value
                    elif field == "UnixStartTime":
                        try:
                            unix_start_time = int(value)
                        except ValueError:
                            print(f"[WARNING] Failed to cast to int {field} with value '{value}' in line {line_num}. "
                                  f"Ignoring it.")
                    elif field == "TimeZone":
                        # this field is ignored in current versions of SWF
                        pass
                    elif field == "TimeZoneString":
                        time_zone_string = value
                    elif field == "StartTime":
                        start_time = parse_time(value)
                    elif field == "EndTime":
                        end_time = parse_time(value)
                    elif field == "MaxJobs":
                        try:
                            max_jobs = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxRecords":
                        try:
                            max_records = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxNodes":
                        try:
                            max_nodes = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxProcs":
                        try:
                            max_procs = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "Preemption":
                        if value.casefold() in ["yes", "no", "double", "ts"]:
                            preemption = value
                        else:
                            print(f"[WARNING] Value '{value}' in line {line_num} for Preemption field is incorrect. "
                                  f"Ignoring it")
                    elif field == "Queues":
                        queues = value
                    elif field == "Queue":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        pass
                    elif field == "MaxQueues":
                        try:
                            max_queues = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "Partitions":
                        partitions = value
                    elif field == "Partition":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        pass
                    elif field == "MaxPartitions":
                        try:
                            max_partitions = int(value)
                        except ValueError:
                            print(f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}."
                                  f"Ignoring it.")
                    elif field == "Note":
                        note = value
                    else:
                        print(f"[WARNING] Unrecognized field {field} in the header in line {line_num}. Ignoring it.")

                else:
                    # this line is not a comment! possible a job
                    # we try to split with multiple tokens. SWF requires white spaces to split lines, but here we support comma as well.
                    try:
                        job_id, submit_time, wait_time, runtime, alloc_cpus, cpus, mem, req_cpus, req_time, req_mem, \
                            status, uid, gid, app_id, qos, partition, previous, think_time = line.split()
                    except ValueError:
                        print(f"[WARNING] Failed to split using white spaces job line {line_num}. Trying with comma...")
                        try:
                            job_id, submit_time, wait_time, runtime, alloc_cpus, cpus, mem, req_cpus, req_time, req_mem, \
                                status, uid, gid, app_id, qos, partition, previous, think_time = line.split(",")
                        except ValueError:
                            print(f"[WARNING] Couldn't split job line {line_num} with either white spaces or comma. "
                                  f"Check if there are 18 fields. Ignoring it... ")
                            continue # go to the next iteration, that is, the next line
                    # at this point we should have the unpacked values
                    job_list.append(
                        Job(job_id, submit_time, wait_time, runtime, alloc_cpus, cpus, mem, req_cpus, req_time,
                            req_mem, status, uid, gid, app_id, qos, partition, previous, think_time))
                    # computing stats
                    avg_cpu += int(alloc_cpus)
                    var_cpu += int(alloc_cpus)**2
                    if uid not in unique_uid:
                        unique_uid.append(uid)
                    if gid not in unique_gid:
                        unique_gid.append(gid)

        # we finish here the computation of stats
        avg_cpu /= len(job_list)
        var_cpu = var_cpu / len(job_list) - avg_cpu**2
        avg_job_rate = len(job_list) / (end_time - start_time).seconds

        return cls(swf_version, computer_name, location, acknowledge, conversion, max_jobs, max_records, preemption,
                 unix_start_time, time_zone_string, start_time, end_time, max_nodes, max_procs, max_runtime, max_memory,
                    allow_overuse, max_queues, queues, queue, max_partitions, partitions, partition, note, job_list,
                   input_file, avg_cpu, var_cpu, avg_job_rate, unique_uid, unique_gid)

    @property
    def swf_version(self):
        return self._swf_version

    @swf_version.setter
    def swf_version(self, version):
        self._swf_version = version

    @property
    def computer(self):
        return self._computer

    @computer.setter
    def computer(self, computer):
        self._computer = computer

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, location):
        self._location = location

    @property
    def acknowledge(self):
        return self._acknowledge

    @acknowledge.setter
    def acknowledge(self, acknowledge):
        self._acknowledge = acknowledge

    @property
    def conversion(self):
        return self._conversion

    @conversion.setter
    def conversion(self, conversion):
        self._conversion = conversion

    @property
    def max_jobs(self):
        return self._max_jobs

    @max_jobs.setter
    def max_jobs(self, max_jobs):
        self._max_jobs = max_jobs

    @property
    def max_records(self):
        return self._max_records

    @max_records.setter
    def max_records(self, max_records):
        self._max_records = max_records

    @property
    def preemption(self):
        return self._preemption

    @preemption.setter
    def preemption(self, preemption):
        self._preemption = preemption

    @property
    def unix_start_time(self):
        return self._unix_start_time

    @unix_start_time.setter
    def unix_start_time(self, unix_start_time):
        self._unix_start_time = unix_start_time

    @property
    def time_zone_string(self):
        return self._time_zone_string

    @time_zone_string.setter
    def time_zone_string(self, time_zone_string):
        self._time_zone_string = time_zone_string

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        self._start_time = parse_time(start_time)
    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    def end_time(self, end_time):
        self._end_time = parse_time(end_time)

    @property
    def max_nodes(self):
        return self._max_nodes

    @max_nodes.setter
    def max_nodes(self, max_nodes):
        self._max_nodes = max_nodes

    @property
    def max_procs(self):
        return self._max_procs

    @max_procs.setter
    def max_procs(self, max_procs):
        self._max_procs = max_procs

    @property
    def max_runtime(self):
        return self._max_runtime

    @max_runtime.setter
    def max_runtime(self, max_runtime):
        self._max_runtime = max_runtime

    @property
    def max_memory(self):
        return self._max_memory

    @max_memory.setter
    def max_memory(self, max_memory):
        self._max_memory = max_memory

    @property
    def allow_overuse(self):
        return self._allow_overuse

    @allow_overuse.setter
    def allow_overuse(self, allow_overuse):
        self._allow_overuse = allow_overuse

    @property
    def max_queues(self):
        return self._max_queues

    @max_queues.setter
    def max_queues(self, max_queues):
        self._max_queues = max_queues

    @property
    def queues(self):
        # TODO how to set multiple lines of comment here
        return self._queues

    @queues.setter
    def queues(self, queues):
        # TODO how to set multiple lines of comment here
        self._queues = queues

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, queue):
        self._queue = queue

    @property
    def max_partitions(self):
        return self._max_partitions

    @max_partitions.setter
    def max_partitions(self, max_partitions):
        self._max_partitions = max_partitions

    @property
    def partitions(self):
        return self._partition

    @partitions.setter
    def partitions(self, partition):
        self._partition = partition

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, note):
        self._note = note

    @property
    def job_list(self):
        return self._job_list

    @job_list.setter
    def job_list(self, new_list):
        """
        This function alters the whole job list of the class
        :param new_list: the job list
        :return:
        """
        self._job_list = new_list[:] # copy the whole list

    def add_job(self, new_job: Job, position: int = None) -> None:
        """
        Add a job to the list
        :param new_job: Job to be added to the trace
        :param position: position in the list where the job will be placed. If nothing is passed, the job is appended.
        :return: None
        """

        if self.allow_overuse == "No":
            # check if the job is within the specs of the swf
            pass
        if position is not None:
            self._job_list.insert(new_job, position)
            # adjust indexes of the following jobs
        else:
            self._job_list.append(new_job)

        # adjust the total number of jobs
        self.max_jobs += 1

    def to_bsc_swf(self, output_filename: str) -> None:
        """
        Outputs a file without headers and with commas for each of the job's attributes.
        :param output_filename: name of the (bsc) swf file
        :return: None
        """
        try:
            with open(output_filename, "w") as file:
                for job in self._job_list:
                    file.write(job.to_bsc_swf())
        except FileNotFoundError:
            print(f"[FATAL] Couldn't create file '{output_filename}'.")

    def user_stats(self):
        """
        Prints statistics about the users in this workload.
        :return:
        """
        print(f"total users: {len(self._unique_uid)}\nunique uids: {self._unique_uid}\n"
              f"total groups: {len(self._unique_gid)}\nunique gids: {self._unique_gid}\n")

    def job_stats(self):
        """
        Prints basic statistics about the jobs in this workload
        :return:
        """
        print(f"number of jobs: {len(self._job_list)}\naverage allocated cpus: {self._avg_cpu}\nvariance of allocated cpus: {self._var_cpu}\n"
              f"rate of submission (jobs/s):  {self._avg_job_rate}")